theory int_int
  imports nat_nat pre_merge_int "../mml/newton"
begin

notation
  uminus ("-\<^sub>\<H> _" [81] 80)

notation
  minus (infixl "-\<^sub>\<H>" 65)

term "less_eq_int"
(*Dirichlet's Theorem*)
text_raw \<open>\DefineSnippet{h2sZ}{\<close>
function h2sZ :: "int \<Rightarrow> Set" ("h2s\<^sub>\<int>'(_')")where
  "x \<ge> 0 \<Longrightarrow> h2s\<^sub>\<int>(x) =\<^sub>\<S> h2s\<^sub>\<nat>(nat(x))"
| "x < 0 \<Longrightarrow> h2s\<^sub>\<int>(x) =\<^sub>\<S> -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(x)))"
text_raw \<open>}%EndSnippet\<close>
proof(auto)
  fix P and x :: int
  assume a: "\<And>xa. 0 \<le> xa \<Longrightarrow> x =\<^sub>\<H> xa \<Longrightarrow> P"
  assume b: "\<And>xa. xa < 0 \<Longrightarrow> x =\<^sub>\<H> xa \<Longrightarrow> P"
  have "x\<ge>0 \<or> x <0" by linarith
  then show P using a b by blast
qed

notation h2sZ ("h2s\<^sub>\<int>")

termination
  apply standard 
  apply auto done
text_raw \<open>\DefineSnippet{s2hZ}{\<close>
function s2hZ :: "Set \<Rightarrow> int" ("s2h\<^sub>\<int>'(_')")where
  "\<not>x is Integer \<Longrightarrow> s2h\<^sub>\<int>(x) =\<^sub>\<H> undefined"
| "x is natural \<Longrightarrow> s2h\<^sub>\<int>(x) =\<^sub>\<H> int(s2h\<^sub>\<nat>(x))"
| "x is Integer & not x is natural \<Longrightarrow> s2h\<^sub>\<int>(x) =\<^sub>\<H> -\<^sub>\<H> (int(s2h\<^sub>\<nat>(-\<^sub>\<S>\<^sup>\<real> x))) "
text_raw \<open>}%EndSnippet\<close>
proof(auto)
  fix x assume A: "\<not> x be integer" and [ty]:"x be natural" 
  have "x be Integer" by mty auto
  thus " undefined =\<^sub>\<H> int(s2h\<^sub>\<nat>(x))" using A by simp
qed

notation s2hZ ("s2h\<^sub>\<int>")

termination
  apply standard 
  apply auto done

theorem h2sZprop: 
  "h2s\<^sub>\<int>(i) is Integer"
proof(cases "i \<ge>0")
  case True
    hence [ty]: "h2s\<^sub>\<int>(i) is Nat" using h2sZ.psimps I2Mprop by auto
    show "h2s\<^sub>\<int>(i) is Integer" by mty auto
  next 
  case False      
    hence "i <0" by linarith
    hence A1:"h2s\<^sub>\<int>(i) = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i)))" using h2sZ.psimps[of i] by auto
    have [ty]: "h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i))) is Nat" using  I2Mprop by auto 
    show "h2s\<^sub>\<int>(i) is Integer" using A1 by mty auto
  qed

theorem h2sZpropE:
  "h2s\<^sub>\<int>(i) is Element-of INT"
proof-
  have "h2s\<^sub>\<int>(i) is Integer" using h2sZprop by auto
  then show "h2s\<^sub>\<int>(i) is Element-of INT" using int_1_def_2 Element(6) by mty auto
qed

theorem MIMZprop:
   "i is Integer implies h2s\<^sub>\<int>(s2h\<^sub>\<int>(i)) =\<^sub>\<S> i"
proof
  assume [ty]:"i is Integer"
  have [ty]: "i be set" using all_set by auto
  show "h2s\<^sub>\<int>(s2h\<^sub>\<int>(i)) = i"
  proof(cases "i is natural")
    case [ty]: True
    then show "h2s\<^sub>\<int>(s2hZ (i)) = i" using MIMprop by auto 
  next case F[ty]: False
    hence A1: "s2hZ (i) =\<^sub>\<H>  -\<^sub>\<H> (int(s2h\<^sub>\<nat>(-\<^sub>\<S>\<^sup>\<real> i)))" by auto
    have [ty]:"-\<^sub>\<S>\<^sup>\<real> i is Nat" using F by mty auto 
    show ?thesis
    proof(cases "s2hZ (i)<0")
      case True 
        hence " h2s\<^sub>\<int>(s2hZ (i)) = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<nat>(nat(-\<^sub>\<H>( -\<^sub>\<H> (int(s2h\<^sub>\<nat>(-\<^sub>\<S>\<^sup>\<real> i))))))" using A1 by auto
        then show "h2s\<^sub>\<int>(s2hZ (i)) =i " using real_0_th_20a MIMprop by mty auto 
      next
        case False 
          hence "s2hZ (i) =\<^sub>\<H> 0" by auto
          hence "h2s\<^sub>\<nat>(s2h\<^sub>\<nat>(-\<^sub>\<S>\<^sup>\<real> i)) = {}" using A1 MIMprop by auto
          hence "(-\<^sub>\<S>\<^sup>\<real> i) = {}" using MIMprop by auto
          hence " i = {}" using real_0_th_18 real_0_th_20a[of i] by mty auto                
          then show "h2s\<^sub>\<int>(s2hZ (i)) = i" by mty auto
    qed
  qed
qed

theorem h2s_less_zero:
  "i <0 implies h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i))) <> {} & h2s\<^sub>\<int>(i) = [{},h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i)))]" 
proof(standard, standard)
  assume I: "i <0" 
      have [ty]: "h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i))) is Nat" using I2Mprop by auto
      show I1: "h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i))) <> {}"
      proof
        assume "h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i))) ={}"
        hence "s2h\<^sub>\<nat>(h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i)))) =\<^sub>\<H> s2h\<^sub>\<nat>(h2s\<^sub>\<nat>(0))" by auto
        then show False using I IMIprop  by auto
      qed
      then show  "h2s\<^sub>\<int>(i) = [{},h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i)))]" using real_0_th_17 I by mty auto
qed
theorem IMIZprop:
  "s2h\<^sub>\<int>(h2s\<^sub>\<int>(i)) =\<^sub>\<H> i"
proof(cases "i \<ge> 0")
  case T: True
    hence A1: "h2s\<^sub>\<int>(i) = h2s\<^sub>\<nat>(nat(i))" by auto
    hence [ty]: "h2s\<^sub>\<nat>(nat(i)) is Nat"
       using I2Mprop by auto 
    hence "s2h\<^sub>\<int>(h2s\<^sub>\<int>(i)) =\<^sub>\<H> int(s2h\<^sub>\<nat>(h2s\<^sub>\<nat>(nat(i))))" using s2hZ.simps A1 by auto
    then show "s2h\<^sub>\<int>(h2s\<^sub>\<int>(i)) =\<^sub>\<H> i" using IMIprop A1 T by auto
  next 
    case False
      hence I: "i <0" by linarith
      have [ty]: "h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i))) is Nat" using I2Mprop by auto
      have [ty]:"h2s\<^sub>\<int>(i) is Integer" using h2sZprop by auto    
      have A1: "h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i))) <> {} & h2s\<^sub>\<int>(i) = [{},h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i)))]" using  h2s_less_zero I by auto
      hence "not h2s\<^sub>\<int>(i) is natural" by mty auto
      hence "s2hZ (h2s\<^sub>\<int>(i)) =\<^sub>\<H>  -\<^sub>\<H> (int(s2h\<^sub>\<nat>(-\<^sub>\<S>\<^sup>\<real>  h2s\<^sub>\<int>(i))))" using s2hZ.simps(3)[of "h2s\<^sub>\<int>(i)"]  by mty auto
      also have "... =\<^sub>\<H> -\<^sub>\<H> (-\<^sub>\<H>(i))" using IMIprop A1 I real_0_th_17[OF ] by mty auto
      finally show "s2h\<^sub>\<int>(h2s\<^sub>\<int>(i)) =\<^sub>\<H> i" by auto
    qed

text_raw \<open>\DefineSnippet{beIsoTInt}{\<close>
theorem beIsoT_Int: 
  "beIsoT(h2s\<^sub>\<int>,s2h\<^sub>\<int>,Integer)"
text_raw \<open>}%EndSnippet\<close>
  unfolding beIsoT_def 
proof(intro conjI)
  show "\<forall>\<^sub>Ly. s2h\<^sub>\<int>(h2s\<^sub>\<int>(y)) =\<^sub>\<H> y" using IMIZprop by auto
  show "\<forall>x:Integer. h2s\<^sub>\<int>(s2h\<^sub>\<int>(x)) = x" using MIMZprop by mty auto
  show "\<forall>\<^sub>Ly. h2s\<^sub>\<int>(y) be Integer"  using h2sZprop by auto
qed
text_raw \<open>\DefineSnippet{beIsoSINT}{\<close>
theorem beIsoS_INT:
  "beIsoS(h2s\<^sub>\<int>,s2h\<^sub>\<int>,INT)"
text_raw \<open>}%EndSnippet\<close>
  by(auto simp add: int_1_def_2 all_set beIsoS_Type_Set(1)[OF _ beIsoT_Int])



theorem positive_add_positive:
   "i \<ge> 0 & j \<ge> 0 implies h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)"
proof
  assume T1: "i \<ge> 0 & j \<ge> 0"
  hence E1: "h2s\<^sub>\<int>(i) = h2s\<^sub>\<nat>(nat(i))" "h2s\<^sub>\<int>(j) = h2s\<^sub>\<nat>(nat(j))" by auto
  have R1[ty]: "h2s\<^sub>\<nat>(nat(i)) is Nat" 
             "h2s\<^sub>\<nat>(nat(j)) is Nat"  using I2Mprop by auto
  have A1: "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<nat>(nat(i+\<^sub>\<H>j))" using T1 by auto
  have "nat(i+\<^sub>\<H>j) =\<^sub>\<H> nat(i) +\<^sub>\<H> nat(j)" using T1 by auto
  hence "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = (h2s\<^sub>\<nat>(nat(i)) +\<^sub>\<S>\<^sup>\<nat> h2s\<^sub>\<nat>(nat(j)))" using A1 AddI by auto
  then show "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" 
    using int_1_def_3[of "h2s\<^sub>\<nat>(nat(i))" "h2s\<^sub>\<nat>(nat(j))",simplified] E1 by mty auto
qed 

theorem positive_add_negative:
  "i \<ge> 0 & j < 0 implies h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)"
proof
  let ?I="h2s\<^sub>\<nat>(nat(i))"
  let ?MJ="h2s\<^sub>\<nat>(nat(-\<^sub>\<H> j))"
  let ?IJ="h2s\<^sub>\<nat>(nat(i+\<^sub>\<H>j))"
  let ?MIJ="h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i+\<^sub>\<H>j)))"
  have R1[ty]: "?I is Nat" "?MJ is Nat" "?IJ is Nat" "?MIJ is Nat"
    using I2Mprop by auto
  assume A1: "i \<ge> 0 & j < 0"
  hence A2: "h2s\<^sub>\<int>(i) = ?I" "h2s\<^sub>\<int>(-\<^sub>\<H> j) = ?MJ" "?MJ<>{}"
        "h2s\<^sub>\<int>(j) = [{},?MJ]" using h2s_less_zero by auto
  have  A[ty]: "h2s\<^sub>\<int>(j) is non natural" "h2s\<^sub>\<int>(j) is Integer" using h2sZprop[of j] A2 by mty auto
  
  have [ty]: "h2s\<^sub>\<int>(i) be Element-of REAL\<^bsup>\<ge>0\<^esup>" using A2 by mty auto
  have [ty]:"\<not> h2s\<^sub>\<int>(j) be Element-of REAL\<^bsup>\<ge>0\<^esup>" by mty auto
  have [ty]:"?MJ be Element-of REAL\<^bsup>\<ge>0\<^esup>" by mty auto
  have [ty]:"h2s\<^sub>\<int>(i) be Mreal" by mty auto
  have [ty]:"h2s\<^sub>\<int>(j) be Mreal" by mty auto
  have A3: "h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j) = ?I --R ?MJ" using real_0_def_4[of "h2s\<^sub>\<int>(i)" "h2s\<^sub>\<int>(j)",simplified]
    A2 A by mty auto
  show "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)"
  proof(cases "i+\<^sub>\<H>j\<ge>0")
    case T1: True
      hence A4: "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = ?IJ" by auto
      have "nat(i+\<^sub>\<H>j) =\<^sub>\<H> i+\<^sub>\<H>j " using T1 by auto
      have "i+\<^sub>\<H>j +\<^sub>\<H> (-\<^sub>\<H> j) =\<^sub>\<H> i" by auto
      hence "?I = ?IJ +\<^sub>\<S>\<^sup>\<real> ?MJ" using A1 T1 positive_add_positive[of "i+\<^sub>\<H>j" "-\<^sub>\<H> j"] A2 by auto
      hence "?I = ?MJ +\<^sub>\<S>\<^sup>\<nat> ?IJ" using Add_6[of ?IJ ?MJ] int_1_def_3[of ?IJ ?MJ] by mty auto
      hence "?I--R ?MJ = ?IJ" using int_0_th_1[of ?I ?IJ ?MJ,simplified] by mty auto
      then show "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" using A4 A3 by auto
  next
    case False
      hence "i+\<^sub>\<H>j <0" by auto
      hence A4: "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = [{}, ?MIJ]" "?MIJ<>{}" "-\<^sub>\<H>(i+\<^sub>\<H>j) \<ge>0" using h2s_less_zero[of "i+\<^sub>\<H>j"] by auto
      have "i+\<^sub>\<H> -\<^sub>\<H> (i+\<^sub>\<H>j) =\<^sub>\<H> -\<^sub>\<H> j" by auto
      hence A5: "?I +\<^sub>\<S>\<^sup>\<real> ?MIJ = ?MJ" using positive_add_positive[of "i" "-\<^sub>\<H> (i+\<^sub>\<H>j)"]  A1 A4(3) by auto
      have R2[ty]: "?MIJ is Integer" "?I is Integer" using R1(4) cl_nat_int by mty auto
      have "?I +\<^sub>\<S>\<^sup>\<real> ?MIJ = ?I +\<^sub>\<S>\<^sup>\<nat> ?MIJ" 
        using int_1_def_3[of ?I ?MIJ] R1 by mty auto
      hence A6: "(?I +\<^sub>\<S>\<^sup>\<nat> ?MIJ) = ?MJ" using A5 by mty auto
      show "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" using int_0_th_3[of ?MJ ?MIJ ?I,OF R1(2,4,1) A6 A4(2)] A4(1) A3 by mty auto   
    qed
  qed

theorem h2sZp:
  "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(j+\<^sub>\<H>i)"
proof(cases "i+\<^sub>\<H>j \<ge>0")
  case True
    hence " h2s\<^sub>\<int>(i +\<^sub>\<H> j) = h2s\<^sub>\<nat>(nat(i +\<^sub>\<H> j))"  "h2s\<^sub>\<int>(j +\<^sub>\<H> i) = h2s\<^sub>\<nat>(nat(j +\<^sub>\<H> i))" "nat(i +\<^sub>\<H> j) =\<^sub>\<H> nat(j +\<^sub>\<H> i)"
      using h2sZ.simps[of "i+\<^sub>\<H>j"] h2sZ.simps[of "j+\<^sub>\<H>i"] by auto 
    then show "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(j+\<^sub>\<H>i)" by auto
  next
    case False
     hence " h2s\<^sub>\<int>(i +\<^sub>\<H> j) = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i +\<^sub>\<H> j)))"  "h2s\<^sub>\<int>(j +\<^sub>\<H> i) = -\<^sub>\<S>\<^sup>\<real>  h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(j +\<^sub>\<H> i)))" "nat(-\<^sub>\<H>(i +\<^sub>\<H> j)) =\<^sub>\<H> nat(-\<^sub>\<H>(j +\<^sub>\<H> i))"
      using h2sZ.simps[of "i+\<^sub>\<H>j"] h2sZ.simps[of "j+\<^sub>\<H>i"] by auto 
    then show "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(j+\<^sub>\<H>i)" by auto
qed

theorem AddIntI:
  "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)"
proof(cases "i \<ge>0")
  case T1: True
  show "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)"
  proof(cases "j \<ge>0")
    case T2: True
       then show "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" using positive_add_positive T1 by auto
     next 
       case F2: False
       then show "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" using positive_add_negative T1 by auto
  qed  
next
  case False
  hence F1: "i <0" by auto
  show "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)"
  proof(cases "j \<ge>0")
    case T2: True
      have [ty]:"h2s\<^sub>\<int>(i) is Integer" "h2s\<^sub>\<int>(j) is Integer" using h2sZprop[of i] h2sZprop[of j] by mty auto
    have "h2s\<^sub>\<int>(j+\<^sub>\<H>i)= h2s\<^sub>\<int>(j) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(i)" using T2 positive_add_negative F1 by auto
        also have "...= h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" using real_0_th_11   by mty auto
      finally show "h2s\<^sub>\<int>(i+\<^sub>\<H>j) =h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" using h2sZp by auto
    next
      case  False
        hence F2:"j <0" by auto
        let ?I = "h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i)))" let ?J = "h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(j)))"  let ?IJ = "h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i+\<^sub>\<H>j)))" 
        have [ty]: "?I is Nat" "?J is Nat"  using I2Mprop h2sZprop by auto
        have F3: "i+\<^sub>\<H>j<0" using F1 F2 by auto
        hence D1: "?IJ <>{}" " h2s\<^sub>\<int>(i+\<^sub>\<H>j) = [{},?IJ]" using h2s_less_zero[of "i+\<^sub>\<H>j"] by auto
        hence D2: "?I <>{}" " h2s\<^sub>\<int>(i) = [{},?I]" "?J <>{}" " h2s\<^sub>\<int>(j) = [{},?J]" 
          using F1 F2 h2s_less_zero by auto+
        hence [ty]: "[{},?I] is Integer" 
             "[{},?J] is Integer" using h2sZprop[of i] h2sZprop[of j] by auto
        have H1: "h2s\<^sub>\<int>(-\<^sub>\<H>(i+\<^sub>\<H>j)) = h2s\<^sub>\<int>(-\<^sub>\<H> i)+\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(-\<^sub>\<H> j)" using  positive_add_positive[of "-\<^sub>\<H> i" "-\<^sub>\<H> j"] F1 F2 by auto
        have H2: "h2s\<^sub>\<int>(-\<^sub>\<H>(i+\<^sub>\<H>j)) = ?IJ" "h2s\<^sub>\<int>(-\<^sub>\<H> j ) = ?J" "h2s\<^sub>\<int>(-\<^sub>\<H> i ) = ?I" using F1 F2 F3 by auto
        hence "?I +\<^sub>\<S>\<^sup>\<real> ?J = ?I +\<^sub>\<S>\<^sup>\<nat> ?J" using int_1_def_3[of ?I ?J,simplified] by mty auto
        hence D3: "?I +\<^sub>\<S>\<^sup>\<nat> ?J = ?IJ" using H1 H2 by auto

       have "[{},?I] +\<^sub>\<S>\<^sup>\<real> [{},?J] = [{}, (?I +\<^sub>\<S>\<^bsup>\<rat>\<^sup>\<ge>\<^sup>0\<^esup> ?J)]"
          using real_0_def_4[of "[{},?I]" "[{},?J]",simplified] using Prat_Preal_add by mty auto
      hence "[{},?I] +\<^sub>\<S>\<^sup>\<real> [{},?J] = [{}, (?I +\<^sub>\<S>\<^sup>\<nat> ?J)]"
          using arytm_3_th_58 by mty auto
        then show "h2s\<^sub>\<int>(i+\<^sub>\<H>j) = h2s\<^sub>\<int>(i) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" using D1 D2 D3 by auto
      qed
qed

theorem AddIntM:
  assumes [ty]: "i is Integer" "j is Integer"
  shows "s2h\<^sub>\<int>(i +\<^sub>\<S>\<^sup>\<real> j) =\<^sub>\<H>  s2h\<^sub>\<int>(i) +\<^sub>\<H> s2h\<^sub>\<int>(j)"
proof-
  have "i +\<^sub>\<S>\<^sup>\<real> j = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i)) +\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(s2h\<^sub>\<int>(j))" using MIMZprop by mty auto
  hence "i +\<^sub>\<S>\<^sup>\<real> j = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i) +\<^sub>\<H> s2h\<^sub>\<int>(j))" using AddIntI[of "s2h\<^sub>\<int>(i)" "s2h\<^sub>\<int>(j)"] by auto
  then show "s2h\<^sub>\<int>(i +\<^sub>\<S>\<^sup>\<real> j) =\<^sub>\<H> s2h\<^sub>\<int>(i) +\<^sub>\<H> s2h\<^sub>\<int>(j)" using IMIZprop[of "s2h\<^sub>\<int>(i) +\<^sub>\<H> s2h\<^sub>\<int>(j)"]  by auto 
qed


theorem h2s_less_eq_zero:
  "i \<le>0 implies h2s\<^sub>\<int>(i) = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(-\<^sub>\<H> i)"
proof
  assume A1: "i \<le>0"
  have [ty]: "h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i))) is Nat" using I2Mprop by auto
  show "h2s\<^sub>\<int>(i) = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(-\<^sub>\<H> i)" 
  proof (cases "i=\<^sub>\<H> 0")
    case True
    hence "h2s\<^sub>\<int>(i) = {}" "h2s\<^sub>\<int>(-\<^sub>\<H> i) = {}" by auto
    then show "h2s\<^sub>\<int>(i) = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(-\<^sub>\<H> i)" using real_0_th_18 by auto
  next
    have [ty]: "h2s\<^sub>\<int>(i) is Integer" using h2sZprop[of i] by mty auto
    case False
    hence A2: "i<0" using A1 by auto
    hence "h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i))) <> {} & h2s\<^sub>\<int>(i) = [{},h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i)))]" 
      using h2s_less_zero by auto
    hence "-\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(i) = h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i)))" using  real_0_th_17 by mty auto
    hence "-\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(i) = h2s\<^sub>\<int>(-\<^sub>\<H>(i))" using  A2 by auto
    then show "h2s\<^sub>\<int>(i) = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(-\<^sub>\<H>(i))" using real_0_th_20a[of "h2s\<^sub>\<int>(i)"]  by mty auto
  qed
qed

theorem h2s_less_eq_zero_1:
  "i \<le>0 implies -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(i) = h2s\<^sub>\<int>(-\<^sub>\<H> i)"
proof
  have [ty]:"h2s\<^sub>\<int>(-\<^sub>\<H> i) is Integer" using h2sZprop by mty auto
  assume "i \<le>0"
  hence "h2s\<^sub>\<int>(i) = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(-\<^sub>\<H>(i))" using h2s_less_eq_zero by auto
  then show "-\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(i) = h2s\<^sub>\<int>(-\<^sub>\<H>(i))"  using real_0_th_20a[of "h2s\<^sub>\<int>(-\<^sub>\<H> i)"]  by mty auto
qed

theorem h2s_less_eq_zero_not_nat:
  "i <0 implies not h2s\<^sub>\<int>(i) is natural"
proof
  assume "i <0"
  hence "h2s\<^sub>\<int>(i) = [{},h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i)))]"  using h2s_less_zero by auto
  then show "not h2s\<^sub>\<int>(i) is natural" by mty auto
qed

theorem h2s_greater_zero:
  "0 < i iff h2s\<^sub>\<int>(i) is natural \<and> h2s\<^sub>\<int>(i)<>{}"
proof(rule iffI3)
  show "0 < i \<longrightarrow> h2s\<^sub>\<int>(i) be natural \<and> h2s\<^sub>\<int>(i) \<noteq> {}"
  proof(rule impI,rule ccontr)
    assume AS: "0<i" "not( h2s\<^sub>\<int>(i) be natural \<and> h2s\<^sub>\<int>(i) \<noteq> {})"
    hence "h2s\<^sub>\<int>(i) \<noteq> {}" using IMIZprop[of i] by mty auto   
    thus False using AS  using I2Mprop  by mty auto
  qed
  assume AS: "h2s\<^sub>\<int>(i) be natural \<and> h2s\<^sub>\<int>(i) \<noteq> {}"
  hence N: "not (i=\<^sub>\<H>0)" by auto
  hence "i \<ge>0" using h2s_less_eq_zero_not_nat[of i] h2sZ.elims[of i] AS by blast
  thus "0 < i" using N by auto
qed


theorem MultIntI:
  "h2s\<^sub>\<int>(i *\<^sub>\<H> j) = h2s\<^sub>\<int>(i) *\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)"
proof(cases "i\<ge>0")
  case T1:True
  hence [ty]: "h2s\<^sub>\<nat>(nat(i)) is Nat" and "h2s\<^sub>\<nat>(nat(i)) =h2s\<^sub>\<int>(i)" using I2Mprop by auto
  hence [ty]: "h2s\<^sub>\<int>(i) is Nat" by auto
  have [ty]: "h2s\<^sub>\<int>(i) is Integer" "h2s\<^sub>\<int>(j) is Integer" 
    using h2sZprop by auto
  show "h2s\<^sub>\<int>(i *\<^sub>\<H> j) = h2s\<^sub>\<int>(i) *\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)"
  proof(cases "j \<ge> 0")
    case T2: True
       have [ty]: "h2s\<^sub>\<nat>(nat(j)) is Nat"  using I2Mprop by auto
       have "h2s\<^sub>\<int>(i *\<^sub>\<H> j) = h2s\<^sub>\<nat>(nat(i *\<^sub>\<H> j))" using T1 T2 by auto
       also have "... =\<^sub>\<H> h2s\<^sub>\<nat>(nat(i) *\<^sub>\<H> nat (j)) " using nat_mult_distrib T1 T2 by auto
       also have "... =\<^sub>\<H> h2s\<^sub>\<nat>(nat(i)) *\<^sub>\<S>\<^sup>\<nat> h2s\<^sub>\<nat>(nat (j)) " using MultI by auto
       also have "... =\<^sub>\<H> h2s\<^sub>\<int>(i) *\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" using T1 T2 int_1_def_5[of "h2s\<^sub>\<int>(i)" "h2s\<^sub>\<int>(j)", simplified] by auto
       finally show "h2s\<^sub>\<int>(i *\<^sub>\<H> j) = h2s\<^sub>\<int>(i) *\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" by auto
     next
       case F2:False
       hence [ty]: "not h2s\<^sub>\<int>(j) is natural" using h2s_less_eq_zero_not_nat by auto
       have "j \<le> 0" using F2 by auto
         hence F3: "i *\<^sub>\<H> j \<le> 0" using T1 mult_nonneg_nonpos by auto 
         have F5: "-\<^sub>\<H> j \<ge>0" using F2 by auto
         hence F4: "i *\<^sub>\<H> (-\<^sub>\<H> j) \<ge>0" using T1 mult_nonneg_nonneg by blast
         hence "h2s\<^sub>\<int>(i *\<^sub>\<H> j) = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(-\<^sub>\<H> (i *\<^sub>\<H> j))" using F3 h2s_less_eq_zero by auto
         also have "... = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(i *\<^sub>\<H> (-\<^sub>\<H> j))" by auto
         also have "... = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<nat>(nat(i *\<^sub>\<H> (-\<^sub>\<H> j)))" using F4 by auto
         also have "... = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<nat>(nat(i) *\<^sub>\<H> nat(-\<^sub>\<H> j)) " using nat_mult_distrib T1 F5 by presburger
         also have "... = -\<^sub>\<S>\<^sup>\<real> (h2s\<^sub>\<nat>(nat(i)) *\<^sub>\<S>\<^sup>\<nat> h2s\<^sub>\<nat>(nat (-\<^sub>\<H> j))) " using MultI by auto
         also have "... = -\<^sub>\<S>\<^sup>\<real> (h2s\<^sub>\<int>(i) *\<^sub>\<S>\<^sup>\<nat> h2s\<^sub>\<nat>(nat (-\<^sub>\<H> j))) " using T1 by auto
         also have "... = -\<^sub>\<S>\<^sup>\<real> (h2s\<^sub>\<int>(i) *\<^sub>\<S>\<^sup>\<nat> (-\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j))) " using F5 h2s_less_eq_zero_1 by auto
         also have "... = h2s\<^sub>\<int>(i) *\<^sub>\<S>\<^sup>\<real>  h2s\<^sub>\<int>(j)" using int_1_def_5[of "h2s\<^sub>\<int>(i)"]  by auto
      finally show "h2s\<^sub>\<int>(i *\<^sub>\<H> j) = h2s\<^sub>\<int>(i) *\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" by auto
  qed
next
  case F1: False
       hence [ty]: "not h2s\<^sub>\<int>(i) is natural" using h2s_less_eq_zero_not_nat by auto
       have I: "i \<le> 0" using F1 by auto
       have F5: "-\<^sub>\<H> i \<ge>0" using F1 by auto
       have [ty]: "h2s\<^sub>\<int>(i) is Integer" "h2s\<^sub>\<int>(j) is Integer" 
        using h2sZprop by auto

       show "h2s\<^sub>\<int>(i *\<^sub>\<H> j) = h2s\<^sub>\<int>(i) *\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)"
       proof(cases "j \<ge> 0")
         case T1: True
           hence [ty]: "h2s\<^sub>\<nat>(nat(j)) is Nat" and "h2s\<^sub>\<nat>(nat(j)) =h2s\<^sub>\<int>(j)" using I2Mprop by auto
           hence [ty]: "h2s\<^sub>\<int>(j) is Nat" by auto
           have F3: "i *\<^sub>\<H> j \<le> 0" using T1 mult_nonpos_nonneg[of i j] I by auto     
           have F4: "(-\<^sub>\<H> i) *\<^sub>\<H> j \<ge>0" using T1 mult_nonneg_nonneg F5 by blast
           hence "h2s\<^sub>\<int>(i *\<^sub>\<H> j) = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(-\<^sub>\<H> (i *\<^sub>\<H> j))" using F3 h2s_less_eq_zero by auto
         also have "... = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(i *\<^sub>\<H> (-\<^sub>\<H> j))" by auto
         also have "... = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<nat>(nat((-\<^sub>\<H> i) *\<^sub>\<H> j))" using F4 by auto
         also have "... = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<nat>(nat(-\<^sub>\<H> i) *\<^sub>\<H> nat(j)) " using nat_mult_distrib T1 F5 by presburger
         also have "... = -\<^sub>\<S>\<^sup>\<real> (h2s\<^sub>\<nat>(nat(-\<^sub>\<H> i)) *\<^sub>\<S>\<^sup>\<nat> h2s\<^sub>\<nat>(nat (j))) " using MultI by auto
         also have "... = -\<^sub>\<S>\<^sup>\<real> (h2s\<^sub>\<nat>(nat(-\<^sub>\<H> i)) *\<^sub>\<S>\<^sup>\<nat> h2s\<^sub>\<int>(j)) " using T1 by auto
         also have "... = -\<^sub>\<S>\<^sup>\<real> ((-\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(i)) *\<^sub>\<S>\<^sup>\<nat> h2s\<^sub>\<int>(j) ) " using F5 h2s_less_eq_zero_1 by auto
         also have "... = h2s\<^sub>\<int>(i) *\<^sub>\<S>\<^sup>\<real>  h2s\<^sub>\<int>(j)" using int_1_def_5[of "h2s\<^sub>\<int>(i)"]  by auto
      finally show "h2s\<^sub>\<int>(i *\<^sub>\<H> j) = h2s\<^sub>\<int>(i) *\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" by auto
    next case G1: False
       hence [ty]: "not h2s\<^sub>\<int>(j) is natural" using h2s_less_eq_zero_not_nat by auto
       have I: "j \<le> 0" using G1 by auto
       have G5: "-\<^sub>\<H> j \<ge>0" using G1 by auto
       have "-\<^sub>\<H> i >0 " "-\<^sub>\<H> j>0" using F1 G1 by auto
       hence G4: "(-\<^sub>\<H> i) *\<^sub>\<H> (-\<^sub>\<H> j) >0"  using mult_pos_pos by blast
       have  "h2s\<^sub>\<int>(i *\<^sub>\<H> j) = h2s\<^sub>\<int>((-\<^sub>\<H> i) *\<^sub>\<H> (-\<^sub>\<H> j))" by auto
       also have "... = h2s\<^sub>\<nat>(nat ((-\<^sub>\<H> i) *\<^sub>\<H> (-\<^sub>\<H> j)))" using G4 by auto
       also have "... = h2s\<^sub>\<nat>(nat(-\<^sub>\<H> i) *\<^sub>\<H> nat (-\<^sub>\<H> j)) " using nat_mult_distrib G5 F5 by presburger 
       also have "... = h2s\<^sub>\<nat>(nat(-\<^sub>\<H> i)) *\<^sub>\<S>\<^sup>\<nat> h2s\<^sub>\<nat>(nat (-\<^sub>\<H> j)) " using MultI by auto
       also have "... = (-\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(i)) *\<^sub>\<S>\<^sup>\<nat> (-\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)) " using h2s_less_eq_zero_1 G5 F5  by auto
       also have "... = h2s\<^sub>\<int>(i) *\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" using int_1_def_5[of "h2s\<^sub>\<int>(i)" "h2s\<^sub>\<int>(j)", simplified] by auto
       finally show "h2s\<^sub>\<int>(i *\<^sub>\<H> j) = h2s\<^sub>\<int>(i) *\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(j)" by auto
     qed
qed

theorem MultIntM:
  assumes [ty]: "i is Integer" "j is Integer"
  shows "s2h\<^sub>\<int>(i *\<^sub>\<S>\<^sup>\<real> j) =\<^sub>\<H>  s2h\<^sub>\<int>(i) *\<^sub>\<H> s2h\<^sub>\<int>(j)"
proof-
  have "i *\<^sub>\<S>\<^sup>\<real> j = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i)) *\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>(s2h\<^sub>\<int>(j))" using MIMZprop by mty auto
  hence "i *\<^sub>\<S>\<^sup>\<real> j = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i) *\<^sub>\<H> s2h\<^sub>\<int>(j))" using MultIntI[of "s2h\<^sub>\<int>(i)" "s2h\<^sub>\<int>(j)"] by auto
  then show "s2h\<^sub>\<int>(i *\<^sub>\<S>\<^sup>\<real> j) =\<^sub>\<H> s2h\<^sub>\<int>(i) *\<^sub>\<H> s2h\<^sub>\<int>(j)" using IMIZprop[of "s2h\<^sub>\<int>(i) *\<^sub>\<H> s2h\<^sub>\<int>(j)"]  by auto 
qed

theorem AssociativeInt:
  assumes [ty]:"i is Integer" "j is Integer" "k is Integer"
  shows "(i *\<^sub>\<S>\<^sup>\<real> j) *\<^sub>\<S>\<^sup>\<real> k = i *\<^sub>\<S>\<^sup>\<real> (j*\<^sub>\<S>\<^sup>\<real> k)"
proof-
  have [ty]: "i*\<^sub>\<S>\<^sup>\<real> j is Integer" "j*\<^sub>\<S>\<^sup>\<real> k is Integer" by mty auto
  have R: "h2s\<^sub>\<int>(s2h\<^sub>\<int>(k)) = k" "h2s\<^sub>\<int>(s2h\<^sub>\<int>(i*\<^sub>\<S>\<^sup>\<real> j)) = i*\<^sub>\<S>\<^sup>\<real> j" using beIsoT_e[OF beIsoT_Int] by auto

   have L: "h2s\<^sub>\<int>(s2h\<^sub>\<int>(i)) = i" "h2s\<^sub>\<int>(s2h\<^sub>\<int>(j*\<^sub>\<S>\<^sup>\<real> k)) = j*\<^sub>\<S>\<^sup>\<real> k" using beIsoT_e[OF beIsoT_Int] by auto  
  have "(i *\<^sub>\<S>\<^sup>\<real> j) *\<^sub>\<S>\<^sup>\<real> k = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i *\<^sub>\<S>\<^sup>\<real> j) *\<^sub>\<H> s2h\<^sub>\<int>(k))" using MultIntI R by auto
  also have "... = h2s\<^sub>\<int>((s2h\<^sub>\<int>(i) *\<^sub>\<H> s2h\<^sub>\<int>(j)) *\<^sub>\<H> s2h\<^sub>\<int>(k))" using MultIntM by auto
  also have "... = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i) *\<^sub>\<H> (s2h\<^sub>\<int>(j) *\<^sub>\<H> s2h\<^sub>\<int>(k)))"
    by (simp add: semiring_normalization_rules(18))
  also have "... = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i) *\<^sub>\<H> s2h\<^sub>\<int>(j *\<^sub>\<S>\<^sup>\<real> k))" using MultIntM by auto
  also have "... =i *\<^sub>\<S>\<^sup>\<real> (j*\<^sub>\<S>\<^sup>\<real> k)" using MultIntI L by auto 
  finally show ?thesis by simp
qed

theorem unitalInt:
  assumes [ty]:"i is Integer"
  shows "i *\<^sub>\<S>\<^sup>\<real> (succ {}) = i" "(succ {}) *\<^sub>\<S>\<^sup>\<real> i = i"
proof-
  let ?S = "succ {}"
  have "s2h\<^sub>\<nat>(succ {}) =\<^sub>\<H> 1" using EqSuc[of "{}"] by mty auto
  hence M: "s2h\<^sub>\<int>(?S) =\<^sub>\<H> 1" using s2hZ.simps[of ?S] by mty auto
  have  "?S is Integer"  by mty auto
  hence L: "h2s\<^sub>\<int>(s2h\<^sub>\<int>(i)) = i" "h2s\<^sub>\<int>(s2h\<^sub>\<int>(?S)) = ?S"  using beIsoT_e[OF beIsoT_Int] by auto  
  hence "(i *\<^sub>\<S>\<^sup>\<real> ?S) = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i) *\<^sub>\<H> 1)" using MultIntI[of "s2h\<^sub>\<int>(i)" "s2h\<^sub>\<int>(?S)"] M by mty auto
  then show "(i *\<^sub>\<S>\<^sup>\<real> ?S) = i" using L by mty auto
  have "(?S *\<^sub>\<S>\<^sup>\<real> i) = h2s\<^sub>\<int>(1 *\<^sub>\<H> s2h\<^sub>\<int>(i))" using MultIntI[of "s2h\<^sub>\<int>(?S)" "s2h\<^sub>\<int>(i)" ] M L by mty auto
  then show "(?S *\<^sub>\<S>\<^sup>\<real> i) = i" using L by mty auto
qed

theorem AbelianInt:
  assumes [ty]:"i is Integer" "j is Integer"
  shows "i *\<^sub>\<S>\<^sup>\<real> j = j *\<^sub>\<S>\<^sup>\<real> i"
proof-
  have  R: "h2s\<^sub>\<int>(s2h\<^sub>\<int>(i)) = i" "h2s\<^sub>\<int>(s2h\<^sub>\<int>(j)) = j" using beIsoT_e[OF beIsoT_Int] by auto
  have "i *\<^sub>\<S>\<^sup>\<real> j = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i) *\<^sub>\<H> s2h\<^sub>\<int>(j))" using MultIntI R by auto
  also have "... = h2s\<^sub>\<int>(s2h\<^sub>\<int>(j) *\<^sub>\<H> s2h\<^sub>\<int>(i))" by (simp add: mult.commute)
  also have "... = j *\<^sub>\<S>\<^sup>\<real> i" using MultIntI R by auto
  finally show ?thesis by auto
qed


theorem distributiveInt:
  assumes [ty]:"i is Integer" "j is Integer" "k is Integer"
  shows "(i +\<^sub>\<S>\<^sup>\<real> j) *\<^sub>\<S>\<^sup>\<real> k = (i *\<^sub>\<S>\<^sup>\<real> k) +\<^sub>\<S>\<^sup>\<real> (j *\<^sub>\<S>\<^sup>\<real> k)"
proof-
  have W[ty]: "i+\<^sub>\<S>\<^sup>\<real> j is Integer" "j*\<^sub>\<S>\<^sup>\<real> k is Integer"  "i*\<^sub>\<S>\<^sup>\<real> k is Integer" by mty auto
  have R: "h2s\<^sub>\<int>(s2h\<^sub>\<int>(k)) = k" "h2s\<^sub>\<int>(s2h\<^sub>\<int>(i+\<^sub>\<S>\<^sup>\<real> j)) = i+\<^sub>\<S>\<^sup>\<real> j" using beIsoT_e[OF beIsoT_Int] by auto

 have L: "h2s\<^sub>\<int>(s2h\<^sub>\<int>(i*\<^sub>\<S>\<^sup>\<real> k)) = i*\<^sub>\<S>\<^sup>\<real> k" "h2s\<^sub>\<int>(s2h\<^sub>\<int>(j*\<^sub>\<S>\<^sup>\<real> k)) = j*\<^sub>\<S>\<^sup>\<real> k" using beIsoT_e[OF beIsoT_Int] by auto  
  
  have "(i +\<^sub>\<S>\<^sup>\<real> j) *\<^sub>\<S>\<^sup>\<real> k = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i +\<^sub>\<S>\<^sup>\<real> j) *\<^sub>\<H> s2h\<^sub>\<int>(k))" using MultIntI R by auto
  also have "... = h2s\<^sub>\<int>((s2h\<^sub>\<int>(i) +\<^sub>\<H> s2h\<^sub>\<int>(j)) *\<^sub>\<H> s2h\<^sub>\<int>(k))" using AddIntM by auto
  also have "... = h2s\<^sub>\<int>((s2h\<^sub>\<int>(i) *\<^sub>\<H> s2h\<^sub>\<int>(k)) +\<^sub>\<H> (s2h\<^sub>\<int>(j) *\<^sub>\<H> s2h\<^sub>\<int>(k)))"
     using int_distrib(1) by auto
  also have "... = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i *\<^sub>\<S>\<^sup>\<real> k) +\<^sub>\<H> s2h\<^sub>\<int>(j *\<^sub>\<S>\<^sup>\<real> k))" using MultIntM by auto
  also have "... = (i *\<^sub>\<S>\<^sup>\<real> k) +\<^sub>\<S>\<^sup>\<real> (j *\<^sub>\<S>\<^sup>\<real> k)" using AddIntI L by auto 
  finally show ?thesis by simp
qed


theorem leqZ:   
  "h2s\<^sub>\<int>(i) is Nat \<longleftrightarrow> i \<ge>0"
proof(rule iffI2)
  show "h2s\<^sub>\<int>(i) is Nat \<longrightarrow> i \<ge>0"
  proof(rule impI, rule ccontr)
    assume A[ty]: "h2s\<^sub>\<int>(i) is Nat"
    have [ty]: "h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i))) is Nat" using I2Mprop by mty auto 
    assume "\<not> 0 \<le> i"
    hence L: "i <0" by linarith
    hence A2: "h2s\<^sub>\<int>(i) = -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i)))" using h2sZ.psimps[of i] by auto
    hence A1:" -\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i))) is Nat" using h2sZ.psimps[of i] A by auto    
    hence A3: "h2s\<^sub>\<int>(i) = {}" using int_1_th0[of "h2s\<^sub>\<nat>(nat(-\<^sub>\<H>(i)))"] A2 by mty auto
    have "s2h\<^sub>\<int>({}) =\<^sub>\<H> s2h\<^sub>\<nat>({})" using  s2hZ.simps by mty auto
    hence "i =\<^sub>\<H> 0" using IMIZprop[of i] s2hZ.simps s2hn.simps A3 by auto
    thus False using L by auto
  qed   
  show "i \<ge>0 \<longrightarrow> (h2s\<^sub>\<int>(i) is Nat)" using h2sZ.psimps I2Mprop by mty auto
qed


theorem zeroZ:   
  "h2s\<^sub>\<int>(i) ={} \<longleftrightarrow> i =\<^sub>\<H> 0"
proof
    assume A1: "h2s\<^sub>\<int>(i) =\<^sub>\<S> {}"
    have "s2h\<^sub>\<int>({}) =\<^sub>\<H> s2h\<^sub>\<nat>({})" by mty auto
    thus "i =\<^sub>\<H> 0" using IMIZprop[of i]  A1 by auto
  next
    assume "i =\<^sub>\<H> 0"
    hence "h2s\<^sub>\<int>(i) =\<^sub>\<H> h2s\<^sub>\<nat>(0)"  by mty auto
    thus "h2s\<^sub>\<int>(i) ={}" by auto
qed

theorem lessZ:   
  "h2s\<^sub>\<int>(i) is Nat & h2s\<^sub>\<int>(i) <>{} \<longleftrightarrow> i > 0" 
using zeroZ[of i] leqZ[of i] by auto


text_raw \<open>\DefineSnippet{intpower}{\<close>
definition int_power :: "Set \<Rightarrow> Set \<Rightarrow> Set" ("_ I^ _"[100,100] 100 ) where
  "int_power \<equiv>  TransformHS2(s2h\<^sub>\<int>,s2h\<^sub>\<nat>,h2s\<^sub>\<int>,(^))"
text_raw \<open>}%EndSnippet\<close>

text_raw \<open>\DefineSnippet{intpowerprop}{\<close>
mtheorem 
  mlet "i be Integer","n be Nat"
  "cluster i I^ n  \<rightarrow> Integer"
text_raw \<open>}%EndSnippet\<close>
 unfolding int_power_def using TransformHS2_typeT[OF beIsoT_Int,of "s2h\<^sub>\<int>" "s2h\<^sub>\<nat>"] by auto

theorem power_MM:
  assumes [ty]: "i be Integer" "n be Nat"
  shows "i I^ n = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i) ^ s2h\<^sub>\<nat>(n))" 
unfolding int_power_def using 
    Transform2_MIT[OF beIsoT_Int beIsoT_Nat,of h2sZ "(^)" "s2h\<^sub>\<int>(i)" "s2h\<^sub>\<nat>(n)"] 
     MIMZprop[of i] MIMprop[of n] by auto

theorem power_MI:
  assumes [ty]: "i be Integer" "n be Nat"
  shows "s2h\<^sub>\<int>(i I^ n) =\<^sub>\<H> s2h\<^sub>\<int>(i) ^ s2h\<^sub>\<nat>(n)" 
  using power_MM[of i n] IMIZprop[of "s2h\<^sub>\<int>(i) ^ s2h\<^sub>\<nat>(n)"] by auto


text_raw \<open>\DefineSnippet{power0}{\<close>
theorem power_01: 
  assumes "i be Integer"
  shows "i I^ 0\<^sub>\<S> =\<^sub>\<S> 1\<^sub>\<S>" and "i I^ 1\<^sub>\<S> =\<^sub>\<S> i"
text_raw \<open>}%EndSnippet\<close>
proof-
  have I: "h2s\<^sub>\<int>(s2h\<^sub>\<int>(i)) = i" "h2s\<^sub>\<nat>(s2h\<^sub>\<nat>({}))={}" using  assms MIMZprop by auto 
  hence "(i I^ {}) = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i) ^ s2h\<^sub>\<nat>({}))" unfolding int_power_def 
    using Transform2_MIT[OF beIsoT_Int beIsoT_Nat,of h2sZ "(^)" "s2h\<^sub>\<int>(i)" "s2h\<^sub>\<nat>({})"] by auto
  then show "i I^ {} = 1\<^sub>\<S>" by auto
    have I: "h2s\<^sub>\<int>(s2h\<^sub>\<int>(i)) = i" "h2s\<^sub>\<nat>(s2h\<^sub>\<nat>(1\<^sub>\<S>))=1\<^sub>\<S>" using  MIMZprop MIMprop[of "1\<^sub>\<S>"] assms by mty auto 
    hence "(i I^ 1\<^sub>\<S>) = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i) ^ s2h\<^sub>\<nat>(1\<^sub>\<S>))" unfolding int_power_def 
    using Transform2_MIT[OF beIsoT_Int beIsoT_Nat,of h2sZ "(^)" "s2h\<^sub>\<int>(i)" "s2h\<^sub>\<nat>(1\<^sub>\<S>)"] I by auto
  then show "i I^ 1\<^sub>\<S> = i" using I  EqSuc[of "{}"] by mty auto

qed

text_raw \<open>\DefineSnippet{powersum}{\<close>
theorem power_sum: 
  assumes "i be Integer" and "n1 be Nat" and "n2 be Nat"
  shows "i I^ (n1 +\<^sub>\<S>\<^sup>\<nat> n2) =\<^sub>\<S> (i I^ n1) *\<^sub>\<S>\<^sup>\<real> (i I^ n2)"
text_raw \<open>}%EndSnippet\<close>
proof-
  have [ty]:"n1 be Nat" "n2 be Nat" using assms by auto
  have N:  "h2s\<^sub>\<nat>(s2h\<^sub>\<nat>(n1))=n1" "h2s\<^sub>\<nat>(s2h\<^sub>\<nat>(n2))=n2" using MIMprop[of n1] MIMprop[of n2] assms by auto
  have I: "h2s\<^sub>\<int>(s2h\<^sub>\<int>(i)) = i" "h2s\<^sub>\<nat>(s2h\<^sub>\<nat>(n1 +\<^sub>\<S>\<^sup>\<nat> n2))=n1 +\<^sub>\<S>\<^sup>\<nat> n2" using MIMprop[of "n1 +\<^sub>\<S>\<^sup>\<nat> n2"] MIMZprop[of i] assms by mty auto 
  hence "(i I^ (n1 +\<^sub>\<S>\<^sup>\<nat> n2)) = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i) ^ s2h\<^sub>\<nat>(n1 +\<^sub>\<S>\<^sup>\<nat> n2))" unfolding int_power_def 
    using Transform2_MIT[OF beIsoT_Int beIsoT_Nat,of h2sZ "(^)" "s2h\<^sub>\<int>(i)" "s2h\<^sub>\<nat>(n1 +\<^sub>\<S>\<^sup>\<nat> n2)"] by auto
  also have "... = h2s\<^sub>\<int>(s2h\<^sub>\<int>(i) ^ (s2h\<^sub>\<nat>(n1)+\<^sub>\<H> s2h\<^sub>\<nat>(n2)))" using AddM by auto
  also have "... = h2s\<^sub>\<int>((s2h\<^sub>\<int>(i) ^ s2h\<^sub>\<nat>(n1)) *\<^sub>\<H> (s2h\<^sub>\<int>(i) ^ s2h\<^sub>\<nat>(n2)))" by (simp add:power_add)
  also have "... = h2s\<^sub>\<int>((s2h\<^sub>\<int>(i) ^ s2h\<^sub>\<nat>(n1))) *\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>((s2h\<^sub>\<int>(i) ^ s2h\<^sub>\<nat>(n2)))" using MultIntI  by auto
  also have "... = (i I^ n1) *\<^sub>\<S>\<^sup>\<real> h2s\<^sub>\<int>((s2h\<^sub>\<int>(i) ^ s2h\<^sub>\<nat>(n2)))"  unfolding int_power_def 
    using Transform2_MIT[OF beIsoT_Int beIsoT_Nat,of h2sZ "(^)" "s2h\<^sub>\<int>(i)" "s2h\<^sub>\<nat>(n1)"] N I by auto
  also have "... = (i I^ n1) *\<^sub>\<S>\<^sup>\<real> (i I^ n2)"  unfolding int_power_def 
    using Transform2_MIT[OF beIsoT_Int beIsoT_Nat,of h2sZ "(^)" "s2h\<^sub>\<int>(i)" "s2h\<^sub>\<nat>(n2)"] N I by auto
  finally show "i I^ (n1 +\<^sub>\<S>\<^sup>\<nat> n2) = (i I^ n1) *\<^sub>\<S>\<^sup>\<real> (i I^ n2)" by auto
qed



text_raw \<open>\DefineSnippet{fermat3}{\<close>
theorem Fermat_3:
  "\<forall>x:Integer. \<forall>y:Integer. \<forall>z:Integer. 
     (x I^ 3\<^sub>\<S>) +\<^sub>\<S>\<^sup>\<real> (y I^3\<^sub>\<S>) = z I^ 3\<^sub>\<S> \<longrightarrow> x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z =\<^sub>\<S> 0\<^sub>\<S>"
text_raw \<open>}%EndSnippet\<close>
proof(intro ballI impI)
  fix x y z assume [ty]:"x be Integer" "y be Integer" "z be Integer" and 
    A1: "(x I^ 3\<^sub>\<S>) +\<^sub>\<S>\<^sup>\<real> (y I^3\<^sub>\<S>) = z I^ 3\<^sub>\<S>"
  have [ty]:"3\<^sub>\<S> is Nat" by mty auto
  have "s2h\<^sub>\<int>(z) ^ 3 =\<^sub>\<H> s2h\<^sub>\<int>(z) ^ s2h\<^sub>\<nat>(3\<^sub>\<S>)" by auto
  also have "... =\<^sub>\<H> s2h\<^sub>\<int>(z I^ 3\<^sub>\<S>)" using power_MI[of _ "3\<^sub>\<S>"]  by auto
  also have "... =\<^sub>\<H> s2h\<^sub>\<int>((x I^ 3\<^sub>\<S>) +\<^sub>\<S>\<^sup>\<real> (y I^3\<^sub>\<S>))" using A1 by auto
  also have "...  =\<^sub>\<H> s2h\<^sub>\<int>(x I^ 3\<^sub>\<S>) +\<^sub>\<H> (s2h\<^sub>\<int>(y I^3\<^sub>\<S>))" 
    using AddIntM[of "x I^ 3\<^sub>\<S>" "y I^ 3\<^sub>\<S>",simplified]  by mty auto
  also have "... =\<^sub>\<H> (s2h\<^sub>\<int>(x) ^ s2h\<^sub>\<nat>(3\<^sub>\<S>)) +\<^sub>\<H> (s2h\<^sub>\<int>(y) ^ s2h\<^sub>\<nat>(3\<^sub>\<S>))" using power_MI[of _ "3\<^sub>\<S>"]  by auto
  also have "... =\<^sub>\<H> (s2h\<^sub>\<int>(x) ^ 3) +\<^sub>\<H> (s2h\<^sub>\<int>(y) ^ 3)" by auto
  finally have A2: "s2h\<^sub>\<int>(x) *\<^sub>\<H> s2h\<^sub>\<int>(y) *\<^sub>\<H> s2h\<^sub>\<int>(z) =\<^sub>\<H> 0" using fermat_3 by auto
 
  have "x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z = h2s\<^sub>\<int>(s2h\<^sub>\<int>((x *\<^sub>\<S>\<^sup>\<real> y) *\<^sub>\<S>\<^sup>\<real> z))" using MIMZprop by mty auto
  also have "... = h2s\<^sub>\<int>(s2h\<^sub>\<int>(x) *\<^sub>\<H> s2h\<^sub>\<int>(y) *\<^sub>\<H> s2h\<^sub>\<int>(z))" using MultIntM by mty auto
  also have "... = 0\<^sub>\<S>" using A2 Zero_One by auto
  finally show  "x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z = 0\<^sub>\<S>" by auto
qed mauto

theorem  Fermat_3dvd:
  "for x,y,z be Integer holds
      for n be Nat st 3\<^sub>\<S> divides n & (x I^ n) +\<^sub>\<S>\<^sup>\<real> (y I^n) = z I^ n
      holds x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z = 0\<^sub>\<S>"
proof(intro ballI impI)
  fix x y z n assume [ty]:"x be Integer" "y be Integer" "z be Integer" "n be Nat" and 
    A1: "3\<^sub>\<S> divides n & (x I^ n) +\<^sub>\<S>\<^sup>\<real> (y I^n) = z I^ n"  
  have [ty]: "x I^ n is Integer" "y I^ n is Integer" by mty auto
  let ?n="s2h\<^sub>\<nat>(n)"
  have "s2h\<^sub>\<nat>(3\<^sub>\<S>)  dvd ?n" using dividesM[of n "3\<^sub>\<S>"] A1 by mty auto
  hence A2: "3 dvd ?n"  by auto
  have "s2h\<^sub>\<int>(z) ^ ?n =\<^sub>\<H>  s2h\<^sub>\<int>(z I^ n)" using power_MI[of _ n]  by auto
  also have "... =\<^sub>\<H> s2h\<^sub>\<int>((x I^ n) +\<^sub>\<S>\<^sup>\<real> (y I^n))" using A1 by auto
  also have "...  =\<^sub>\<H> s2h\<^sub>\<int>(x I^ n) +\<^sub>\<H> (s2h\<^sub>\<int>(y I^n))" 
    using AddIntM[of "x I^ n" "y I^ n",simplified]  by mty auto
  also have "... =\<^sub>\<H> (s2h\<^sub>\<int>(x) ^ ?n) +\<^sub>\<H> (s2h\<^sub>\<int>(y) ^ ?n)" using power_MI[of _ n]  by auto
  finally have A2: "s2h\<^sub>\<int>(x) *\<^sub>\<H> s2h\<^sub>\<int>(y) *\<^sub>\<H> s2h\<^sub>\<int>(z) =\<^sub>\<H> 0" using fermat_mult3[OF _ A2] by auto
 
  have "x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z = h2s\<^sub>\<int>(s2h\<^sub>\<int>((x *\<^sub>\<S>\<^sup>\<real> y) *\<^sub>\<S>\<^sup>\<real> z))" using MIMZprop by mty auto
  also have "... = h2s\<^sub>\<int>(s2h\<^sub>\<int>(x) *\<^sub>\<H> s2h\<^sub>\<int>(y) *\<^sub>\<H> s2h\<^sub>\<int>(z))" using MultIntM by mty auto
  also have "... = 0\<^sub>\<S>" using A2 Zero_One by auto
  finally show  "x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z = 0\<^sub>\<S>" by auto
qed mauto

text_raw \<open>\DefineSnippet{fermat4}{\<close>
theorem Fermat_4:
  "for x,y,z be Integer st (x I^ 4\<^sub>\<S>) +\<^sub>\<S>\<^sup>\<real> (y I^4\<^sub>\<S>) = z I^ 4\<^sub>\<S>
      holds x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z = 0\<^sub>\<S>"
proof(intro ballI impI)
  fix x y z assume [ty]:"x be Integer" "y be Integer" "z be Integer" and 
    A1: "(x I^ 4\<^sub>\<S>) +\<^sub>\<S>\<^sup>\<real> (y I^4\<^sub>\<S>) = z I^ 4\<^sub>\<S>"
  have M1: "s2h\<^sub>\<int>(z) ^ 4 =\<^sub>\<H> (s2h\<^sub>\<int>(x) ^ 4) +\<^sub>\<H> (s2h\<^sub>\<int>(y) ^ 4)" using 
    A1 AddIntM[of "x I^ 4\<^sub>\<S>" "y I^ 4\<^sub>\<S>"] power_MI by mty auto
  hence "s2h\<^sub>\<int>(x) *\<^sub>\<H> s2h\<^sub>\<int>(y) *\<^sub>\<H> s2h\<^sub>\<int>(z) =\<^sub>\<H> 0" using fermat_4 by mty auto
  thus  "x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z = 0\<^sub>\<S>" using  MIMZprop[of "x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z"] MultIntM s2hn.simps by mty auto
qed mauto
text_raw \<open>}%EndSnippet\<close>


theorem  Fermat_4dvd:
  "for x,y,z be Integer holds
      for n be Nat st 4\<^sub>\<S> divides n & (x I^ n) +\<^sub>\<S>\<^sup>\<real> (y I^n) = z I^ n
      holds x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z = 0\<^sub>\<S>"
proof(intro ballI impI)
  fix x y z n assume [ty]:"x be Integer" "y be Integer" "z be Integer" "n be Nat" and 
    A1: "4\<^sub>\<S> divides n & (x I^ n) +\<^sub>\<S>\<^sup>\<real> (y I^n) = z I^ n"  
  let ?n="s2h\<^sub>\<nat>(n)"
  have "s2h\<^sub>\<nat>(4\<^sub>\<S>)  dvd ?n" using dividesM[of n "4\<^sub>\<S>"] A1 by mty auto
  hence A2: "4 dvd ?n" by auto
  have "s2h\<^sub>\<int>(z) ^ ?n =\<^sub>\<H> (s2h\<^sub>\<int>(x) ^ ?n) +\<^sub>\<H> (s2h\<^sub>\<int>(y) ^ ?n)" using 
     power_MI[of _ n] AddIntM[of "x I^ n" "y I^ n"] power_MI[of _ n] A1  by mauto
  hence A3: "s2h\<^sub>\<int>(x) *\<^sub>\<H> s2h\<^sub>\<int>(y) *\<^sub>\<H> s2h\<^sub>\<int>(z) =\<^sub>\<H> 0" using fermat_mult4[OF _ A2] by auto
 
  have "x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z = h2s\<^sub>\<int>(s2h\<^sub>\<int>((x *\<^sub>\<S>\<^sup>\<real> y) *\<^sub>\<S>\<^sup>\<real> z))" using MIMZprop by mty auto
  also have "... = h2s\<^sub>\<int>(s2h\<^sub>\<int>(x) *\<^sub>\<H> s2h\<^sub>\<int>(y) *\<^sub>\<H> s2h\<^sub>\<int>(z))" using MultIntM by mty auto
  also have "... = 0\<^sub>\<S>" using A3 Zero_One by auto
  finally show  "x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z = 0\<^sub>\<S>" by auto
qed mauto

text_raw \<open>\DefineSnippet{fermat34}{\<close>
theorem Fermat_divides_3_4:
  "\<forall>x,y,z:Integer. \<forall>n:Nat.
    (3\<^sub>\<S> divides n \<or> 4\<^sub>\<S> divides n) \<and> (x I^ n) +\<^sub>\<S>\<^sup>\<real> (y I^n) =\<^sub>\<S> z I^n
   \<longrightarrow> x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z =\<^sub>\<S> 0\<^sub>\<S>"
text_raw \<open>}%EndSnippet\<close>
proof(intro ballI impI)
  fix x y z n assume [ty]:"x be Integer" "y be Integer" "z be Integer" "n be Nat" and 
     A1:"(3\<^sub>\<S> divides n or 4\<^sub>\<S> divides n) & 
                     (x I^ n) +\<^sub>\<S>\<^sup>\<real> (y I^n) = z I^ n"
  then show "x *\<^sub>\<S>\<^sup>\<real> y *\<^sub>\<S>\<^sup>\<real> z = 0\<^sub>\<S>"
    using Fermat_4dvd[THEN bspec,THEN bspec,THEN bspec, THEN bspec, of x y z n,simplified] 
    Fermat_3dvd[THEN bspec,THEN bspec,THEN bspec,THEN bspec,of x y z n,simplified] by mty auto
qed mauto

end

